@@ -36,6 +36,7 @@
 #import "Root.h"
 #import "Thread.h"
 #import "CacheAgent.h"
+#import "Config.h"
 #import "LUPrivate.h"
 #import "LUCachedDictionary.h"
 #import <arpa/inet.h>
@@ -208,7 +209,7 @@ static BOOL shutdown_CacheAgent = NO;
 			}
 		}
 
-		all = rootInitGroups.all;
+		all = rootInitGroups;
 		if (all != nil)
 		{
 			if (expired > 0)
@@ -236,8 +237,8 @@ static BOOL shutdown_CacheAgent = NO;
 
 	if (expireRootInitgroups)
 	{
-		[rootInitGroups.all release];
-		rootInitGroups.all = nil;
+		[rootInitGroups release];
+		rootInitGroups = nil;
 	}
 
 	if (allStore[cat].all != nil)
@@ -316,11 +317,92 @@ static BOOL shutdown_CacheAgent = NO;
  * Object creation, initilizations, and general stuff
  */
 
-- (CacheAgent *)init
+- (LUAgent *)initWithArg:(char *)arg
+{
+	return [self init];
+}
+
+- (void)setCacheIsValidated:(BOOL)validate forCategory:(LUCategory)cat
 {
 	int i;
-	char str[128];
+
+	for (i = 0; i < NCACHE; i++)
+	{
+		if (cacheCategory[i] == cat)
+			cacheStore[i].validate = validate;
+	}
+
+	allStore[(unsigned int)cat].validate = validate;
+}
+
+- (void)setCapacity:(unsigned int)max forCategory:(LUCategory)cat
+{
+	int i;
+
+	for (i = 0; i < NCACHE; i++)
+	{
+		if (cacheCategory[i] == cat)
+			cacheStore[i].capacity = max;
+	}
+}
+
+- (void)setTimeToLive:(time_t)timeout forCategory:(LUCategory)cat
+{
+	int i;
+
+	for (i = 0; i < NCACHE; i++)
+	{
+		if (cacheCategory[i] == cat)
+			cacheStore[i].ttl = timeout;
+	}
+}
+
+- (time_t)timeToLiveForCategory:(LUCategory)cat
+{
+	int n;
+
+	n = cacheForCategory[cat];
+	if (n < 0) return 0;
+	return cacheStore[n].ttl;
+}
+
+- (void)setTimeToLive:(time_t)ttl forArray:(LUArray *)array
+{
+	LUDictionary *stamp;
+	int i, len;
+
+	if (array == nil) return;
+
+	len = [array validationStampCount];
+	for (i = 0; i < len; i++)
+	{
+		stamp = [array validationStampAtIndex:i];
+		if (stamp != nil) [stamp setTimeToLive:ttl];
+	}
+}
+
+- (void)setCacheIsEnabled:(BOOL)enabled forCategory:(LUCategory)cat
+{
+	int i;
+
+	for (i = 0; i < NCACHE; i++)
+	{
+		if (cacheCategory[i] == cat)
+			cacheStore[i].enabled = enabled;
+	}
+
+	allStore[(unsigned int)cat].enabled = enabled;
+}
+
+- (CacheAgent *)init
+{
+	int i, j;
+	char str[128], **order;
 	struct timeval now;
+	LUDictionary *global, *config;
+	BOOL gValidation, cValidation, gEnable, cEnable;
+	unsigned int gMax, cMax;
+	time_t gTTL, cTTL;
 
 	if (didInit) return self;
 
@@ -330,7 +412,6 @@ static BOOL shutdown_CacheAgent = NO;
 	lastSweep = now.tv_sec;
 	sweepTime = [self minTimeToLive];
 	if (sweepTime < 60) sweepTime = 60;
-
 	cacheLock = syslock_new(1);
 
 	for (i = 0; i < NCACHE; i++)
@@ -340,8 +421,6 @@ static BOOL shutdown_CacheAgent = NO;
 		[cacheStore[i].cache setBanner:str];
 		cacheStore[i].capacity = (unsigned int)-1;
 		cacheStore[i].ttl = 43200;
-		cacheStore[i].delta = 0;
-		cacheStore[i].freq = 0;
 		cacheStore[i].validate = YES;
 		cacheStore[i].enabled = NO;
 	}
@@ -353,13 +432,53 @@ static BOOL shutdown_CacheAgent = NO;
 		allStore[i].enabled = NO;
 	}
 
-	rootInitGroups.all = nil;
+	rootInitGroups = nil;
 
 	initgroupsUserName = NULL;
 
-	stats = [[LUDictionary alloc] init];
-	[stats setBanner:"CacheAgent statistics"];
-	[stats setValue:"Cache" forKey:"information_system"];
+	global = [configManager configGlobal:configurationArray];
+
+	gValidation = [configManager boolForKey:"ValidateCache" dict:global default:YES];
+	gMax = [configManager intForKey:"CacheCapacity" dict:global default:-1];
+	if (gMax == 0) gMax = (unsigned int)-1;
+	gTTL = (time_t)[configManager intForKey:"TimeToLive" dict:global default:43200];
+
+	gEnable = NO;
+	order = [global valuesForKey:"LookupOrder"];
+	if (order != NULL)
+	{
+		for (i = 0; order[i] != NULL; i++)
+		{
+			if (streq(order[i], "Cache") || streq(order[i], "CacheAgent"))
+				gEnable = YES;
+		}
+	}
+
+	for (i = 0; i < NCATEGORIES; i++)
+	{	
+		config = [configManager configForCategory:i fromConfig:configurationArray];
+		cValidation = [configManager boolForKey:"ValidateCache" dict:config default:gValidation];
+		cMax = [configManager intForKey:"CacheCapacity" dict:config default:gMax];
+		if (cMax == 0) cMax = (unsigned int)-1;
+		cTTL = (time_t)[configManager intForKey:"TimeToLive" dict:config default:gTTL];
+
+		cEnable = gEnable;
+		order = [config valuesForKey:"LookupOrder"];
+		if (order != NULL)
+		{
+			cEnable = NO;
+			for (j = 0; order[j] != NULL; j++)
+			{
+				if (streq(order[j], "Cache") || streq(order[j], "Cache"))
+					cEnable = YES;
+			}
+		}
+
+		[self setCacheIsValidated:cValidation forCategory:(LUCategory)i];
+		[self setCacheIsEnabled:cEnable forCategory:(LUCategory)i];
+		[self setCapacity:cMax forCategory:(LUCategory)i];
+		[self setTimeToLive:cTTL forCategory:(LUCategory)i];
+	}
 
 	return self;
 }
@@ -397,15 +516,12 @@ static BOOL shutdown_CacheAgent = NO;
 		if (allStore[i].all != nil) [allStore[i].all release];
 	}
 
-	[rootInitGroups.all release];
+	[rootInitGroups release];
 
 	if (initgroupsUserName != NULL) freeString(initgroupsUserName);
 	initgroupsUserName = NULL;
 
-	[stats release];
-
-	if (cacheLock != NULL) syslock_free(cacheLock);
-	cacheLock = NULL;
+	syslock_free(cacheLock);
 	
 	system_log(LOG_DEBUG, "Deallocated CacheAgent 0x%08x\n", (int)self);
 
@@ -414,50 +530,28 @@ static BOOL shutdown_CacheAgent = NO;
 	_sharedCacheAgent = nil;
 }
 
-- (const char *)serviceName
-{
-	return "Cache";
-}
-
 - (const char *)shortName
 {
 	return "Cache";
 }
 
-- (LUDictionary *)statistics
+- (BOOL)isValid:(LUDictionary *)item
 {
-	int i;
-	char key[256], str[256];
+	id agent;
 
-	for (i = 0; i < NCACHE; i++)
-	{
-		sprintf(key, "%s size", cacheName[i]);
-		sprintf(str, "%d", [cacheStore[i].cache count]);
-		[stats setValue:str forKey:key];
-	}
-
-	return stats;
-}
-
-- (void)resetStatistics
-{
-	[stats release];
-
-	stats = [[LUDictionary alloc] init];
-	[stats setBanner:"CacheAgent statistics"];
-	[stats setValue:"Cache" forKey:"information_system"];
-}
+	if (item == nil) return NO;
 		
-/*
- * Fetch objects from cache
- */
+	agent = [item agent];
+	if (agent == nil) return NO;
+
+	return [agent isValid: item];
+}
 
 - (LUDictionary *)postProcess:(LUDictionary *)item
 	cache:(unsigned int)n
 	key:(char *)key
 {
-	id agent;
-	time_t age, ttl, newttl;
+	time_t age, ttl;
 	unsigned int hits;
 
 	if (item == nil) return nil;
@@ -465,14 +559,7 @@ static BOOL shutdown_CacheAgent = NO;
 
 	if (cacheStore[n].validate)
 	{
-		agent = [item agent];
-		if (agent == nil)
-		{
-			[cacheStore[n].cache removeObject:item];
-			return nil;
-		}
-
-		if (![agent isValid:item])
+		if (![self isValid:item])
 		{
 			[cacheStore[n].cache removeObject:item];
 			return nil;
@@ -489,16 +576,6 @@ static BOOL shutdown_CacheAgent = NO;
 	hits = [item cacheHit];
 	if (cacheStore[n].validate) [item resetAge];
 
-	if ((cacheStore[n].freq > 0) &&
-		((hits % cacheStore[n].freq) == 0))
-	{
-		newttl = ttl + cacheStore[n].delta;
-
-		/* check for wrap-around */
-		if (newttl < ttl) newttl = (time_t)-1;
-		[item setTimeToLive:newttl];
-	}
-
 	/* Retain the object here.  Caller must release. */
 	[item retain];
 	return item;
@@ -526,7 +603,6 @@ static BOOL shutdown_CacheAgent = NO;
 		if (agent == nil) return NO;
 		if (![agent isValid:stamp]) return NO;
 	}
-
 	return YES;
 }
 
@@ -567,21 +643,6 @@ static BOOL shutdown_CacheAgent = NO;
 	return nil;
 }
 
-- (void)setTimeToLive:(time_t)ttl forArray:(LUArray *)array
-{
-	LUDictionary *stamp;
-	int i, len;
-
-	if (array == nil) return;
-
-	len = [array validationStampCount];
-	for (i = 0; i < len; i++)
-	{
-		stamp = [array validationStampAtIndex:i];
-		if (stamp != nil) [stamp setTimeToLive:ttl];
-	}
-}
-
 - (void)addArray:(LUArray *)array
 {
 	LUDictionary *stamp;
@@ -602,22 +663,20 @@ static BOOL shutdown_CacheAgent = NO;
 	[self setTimeToLive:ttl forArray:array];
 
 	syslock_lock(cacheLock);
-
 	if (allStore[cat].all != nil) [allStore[cat].all release];
 	allStore[cat].all = [array retain];
-
 	syslock_unlock(cacheLock);
 }
 
 - (LUArray *)initgroupsForUser:(char *)name
 {
 	LUArray *all;
-	
+
 	syslock_lock(cacheLock);
 
 	if (streq(name, "root"))
 	{
-		if (rootInitGroups.all == nil)
+		if (rootInitGroups == nil)
 		{
 			syslock_unlock(cacheLock);
 			return nil;
@@ -625,20 +684,20 @@ static BOOL shutdown_CacheAgent = NO;
 
 		if (!allStore[(unsigned int)LUCategoryInitgroups].validate)
 		{
-			[rootInitGroups.all retain];
+			[rootInitGroups retain];
 			syslock_unlock(cacheLock);
-			return rootInitGroups.all;
+			return rootInitGroups;
 		}
 
-		if ([self isArrayValid:rootInitGroups.all])
+		if ([self isArrayValid:rootInitGroups])
 		{
-			[rootInitGroups.all retain];
+			[rootInitGroups retain];
 			syslock_unlock(cacheLock);
-			return rootInitGroups.all;
+			return rootInitGroups;
 		}
 
-		[rootInitGroups.all release];
-		rootInitGroups.all = nil;
+		[rootInitGroups release];
+		rootInitGroups = nil;
 		syslock_unlock(cacheLock);
 		return nil;
 	}
@@ -702,8 +761,8 @@ static BOOL shutdown_CacheAgent = NO;
 
 	if (streq(name, "root"))
 	{
-		if (rootInitGroups.all != nil) [rootInitGroups.all release];
-		rootInitGroups.all = [groups retain];
+		if (rootInitGroups != nil) [rootInitGroups release];
+		rootInitGroups = [groups retain];
 		syslock_unlock(cacheLock);
 		return;
 	}
@@ -1006,9 +1065,9 @@ static BOOL shutdown_CacheAgent = NO;
 
 	if (item == nil) return;
 
+	cat = [item category];
 	syslock_lock(cacheLock);
 
-	cat = [item category];
 	switch (cat)
 	{
 		case LUCategoryUser:
@@ -1066,16 +1125,24 @@ static BOOL shutdown_CacheAgent = NO;
 	syslock_unlock(cacheLock);
 }
 
+- (void)reset
+{
+	syslock_lock(cacheLock);
+	[configurationArray release];
+	generation = [configManager generation];
+	configurationArray = [configManager config];
+	syslock_unlock(cacheLock);
+
+	[self flushCache];
+}
+
 - (void)flushCache
 {
 	int i;
 
 	syslock_lock(cacheLock);
 
-	for (i = 0; i < NCACHE; i++)
-	{
-		[cacheStore[i].cache empty];
-	}
+	for (i = 0; i < NCACHE; i++) [cacheStore[i].cache empty];
 
 	for (i = 0; i < NCATEGORIES; i++)
 	{
@@ -1089,8 +1156,8 @@ static BOOL shutdown_CacheAgent = NO;
 		}
 	}
 
-	if (rootInitGroups.all != nil) [rootInitGroups.all release];
-	rootInitGroups.all = nil;
+	if (rootInitGroups != nil) [rootInitGroups release];
+	rootInitGroups = nil;
 
 	syslock_unlock(cacheLock);
 }
@@ -1115,8 +1182,8 @@ static BOOL shutdown_CacheAgent = NO;
 
 	if (cat == LUCategoryInitgroups)
 	{
-		if (rootInitGroups.all != nil) [rootInitGroups.all release];
-		rootInitGroups.all = nil;
+		if (rootInitGroups != nil) [rootInitGroups release];
+		rootInitGroups = nil;
 	}
 
 	syslock_unlock(cacheLock);
@@ -1126,123 +1193,18 @@ static BOOL shutdown_CacheAgent = NO;
  * Cache management
  */
 
-- (void)setCapacity:(unsigned int)max forCategory:(LUCategory)cat
-{
-	int i;
-
-	for (i = 0; i < NCACHE; i++)
-	{
-		if (cacheCategory[i] == cat)
-			cacheStore[i].capacity = max;
-	}
-}
-
-- (unsigned int)capacityForCategory:(LUCategory)cat
-{
-	int n;
-
-	n = cacheForCategory[cat];
-	if (n < 0) return 0;
-	return cacheStore[n].capacity;
-}
-
-- (void)setTimeToLive:(time_t)timeout forCategory:(LUCategory)cat
-{
-	int i;
-
-	for (i = 0; i < NCACHE; i++)
-	{
-		if (cacheCategory[i] == cat)
-			cacheStore[i].ttl = timeout;
-	}
-}
-
-- (time_t)timeToLiveForCategory:(LUCategory)cat
-{
-	int n;
-
-	n = cacheForCategory[cat];
-	if (n < 0) return 0;
-	return cacheStore[n].ttl;
-}
-
-- (void)setCacheIsValidated:(BOOL)validate forCategory:(LUCategory)cat
-{
-	int i;
-
-	for (i = 0; i < NCACHE; i++)
-	{
-		if (cacheCategory[i] == cat)
-			cacheStore[i].validate = validate;
-	}
-
-	allStore[(unsigned int)cat].validate = validate;
-}
-
-- (BOOL)cacheIsValidatedForCategory:(LUCategory)cat
-{
-	if (cat > NCATEGORIES) return NO;
-	return allStore[(unsigned int)cat].validate;
-}
-
-- (void)setCacheIsEnabled:(BOOL)enabled forCategory:(LUCategory)cat
-{
-	int i;
-
-	for (i = 0; i < NCACHE; i++)
-	{
-		if (cacheCategory[i] == cat)
-			cacheStore[i].enabled = enabled;
-	}
-
-	allStore[(unsigned int)cat].enabled = enabled;
-}
-
 - (BOOL)cacheIsEnabledForCategory:(LUCategory)cat
 {
 	if (cat > NCATEGORIES) return NO;
 	return allStore[(unsigned int)cat].enabled;
 }
 
-- (void)addTimeToLive:(time_t)delta
-	afterCacheHits:(unsigned int)freq
-	forCategory:(LUCategory)cat
-{
-	int i;
-
-	for (i = 0; i < NCACHE; i++)
-	{
-		if (cacheCategory[i] == cat)
-		{
-			cacheStore[i].delta = delta;
-			cacheStore[i].freq = freq;
-		}
-	}
-}
-
-- (time_t)cacheTimeToLiveDeltaForCategory:(LUCategory)cat
-{
-	int n;
-
-	n = cacheForCategory[cat];
-	if (n < 0) return 0;
-	return cacheStore[n].delta;
-}
-
-- (unsigned int)cacheTimeToLiveFrequencyForCategory:(LUCategory)cat
-{
-	int n;
-
-	n = cacheForCategory[cat];
-	if (n < 0) return 0;
-	return cacheStore[n].freq;
-}
-
 - (BOOL)containsObject:(id)obj
 {
 	int i;
 
 	syslock_lock(cacheLock);
+
 	if ([obj isMemberOf:[LUArray class]])
 	{
 		for (i = 0; i < NCATEGORIES; i++)
@@ -1254,7 +1216,7 @@ static BOOL shutdown_CacheAgent = NO;
 			}
 		}
 
-		if (obj == rootInitGroups.all) 
+		if (obj == rootInitGroups)
 		{
 			syslock_unlock(cacheLock);
 			return YES;
@@ -1274,6 +1236,7 @@ static BOOL shutdown_CacheAgent = NO;
 	}
 
 	syslock_unlock(cacheLock);
+
 	return NO;
 }
 
